import os
import sqlite3
from functools import lru_cache
from timeit import default_timer as timer
from typing import Tuple

from openai import ChatCompletion
from tabulate import tabulate

DATABASE = os.environ.get("DATABASE", "nldb.db")


def get_prompt_template() -> str:
    # returns the prompt template as a string, with comments removed
    lines = open("prompt.txt").read().split("\n")
    uncommented_lines = [line for line in lines if not line.strip().startswith("#")]
    return "\n".join(uncommented_lines).strip()


@lru_cache
def text_to_sql(prompt_template: str, question: str) -> Tuple[str, int, float]:
    # uses GPT-3.5 to convert a question into a SQL statement
    prompt_messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": prompt_template % question},
    ]

    start = timer()
    resp = ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=prompt_messages,
        temperature=0,
    )
    gpt_time = timer() - start

    sql = resp["choices"][0]["message"]["content"].strip()
    tokens = resp["usage"]["total_tokens"]

    return (sql, tokens, gpt_time)


@lru_cache
def sql_to_answer(
    sql: str, prompt_template: str, question: str
) -> Tuple[str, str, int, float, float]:
    start = timer()
    (columns, raw_results) = execute_sql(sql)
    execution_time = timer() - start
    formatted_results = tabulate(raw_results, headers=list(columns))
    clean_columns = [column[0].replace("_", " ") for column in columns]
    html_results = tabulate(raw_results, headers=clean_columns, tablefmt="html")

    sql_results_message = (
        "I executed the SQL statement and got the following results:\n\n"
        + formatted_results
    )

    prompt_messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": prompt_template % question},
        {"role": "assistant", "content": sql},
        {"role": "system", "content": sql_results_message},
        {
            "role": "system",
            "content": f"Using these results, answer the question: {question}",
        },
        {
            "role": "system",
            "content": "Don't say what the query was, just answer the question.",
        },
    ]

    gpt_start = timer()
    resp = ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=prompt_messages,
        temperature=0,
    )
    gpt_time = timer() - gpt_start

    answer = resp["choices"][0]["message"]["content"].strip()
    tokens = resp["usage"]["total_tokens"]
    return (html_results, answer, tokens, execution_time, gpt_time)


def execute_sql(query):
    # open the database in read-only mode
    connection = sqlite3.connect(f"file:{DATABASE}?mode=ro", uri=True)
    cursor = connection.cursor()
    cursor.execute(query)
    result = cursor.fetchall()
    columns = list(cursor.description)
    return (columns, result)



import os
from typing import Union

import uvicorn
from fastapi import FastAPI
from fastapi.responses import FileResponse

from nldb.core import get_prompt_template, text_to_sql, sql_to_answer

UVICORN_HOST = os.environ.get("UVICORN_HOST", "0.0.0.0")
UVICORN_PORT = int(os.environ.get("UVICORN_PORT", 8080))

app = FastAPI()


@app.on_event("startup")
def template_loader():
    global prompt_template
    prompt_template = get_prompt_template()


@app.get("/api/ask")
async def ask(q: Union[str, None] = None):
    (sql_statement, tokens_1, time_1) = text_to_sql(prompt_template, q)
    (results, answer, tokens_2, sql_time, time_2) = sql_to_answer(
        sql_statement, prompt_template, q
    )
    tokens = tokens_1 + tokens_2
    return {
        "response": {
            "answer": answer,
            "results": results,
            "sql": sql_statement,
            "timings": [time_1, sql_time, time_2],  # in seconds
            "tokens": tokens,
            "cost": (tokens_1 + tokens_2)
            / 1000
            * 0.002,  # gpt-3.5 is $0.002 per 1000 tokens,
        }
    }


@app.get("/")
async def read_index():
    return FileResponse("index.html")


def serve():
    uvicorn.run(app, host=UVICORN_HOST, port=UVICORN_PORT)



import inspect
import os
import shutil
from typing import Optional

import typer
from nldb.api import serve
from nldb.core import DATABASE


def preflight_checks():
    # check that the prompt.txt file exists
    failures = 0
    if not os.path.exists("prompt.txt"):
        typer.echo("prompt.txt file not found")
        typer.echo("Run `nldb init` to create a new prompt.txt file")
        failures += 1
    if not os.path.exists(DATABASE):
        typer.echo(f"Database file not found: {DATABASE}")
        failures += 1
    if not os.environ.get("OPENAI_API_KEY"):
        typer.echo("OPENAI_API_KEY environment variable not set")
        failures += 1
    if failures > 0:
        raise typer.Exit(code=1)


def init():
    # create a starter Dockerfile, index.html and prompt.py
    package_dir = os.path.dirname(os.path.realpath(__file__))
    for file in ["Dockerfile", "index.html", "prompt.txt"]:
        dest = os.path.join(os.getcwd(), file)
        if os.path.exists(dest):
            typer.echo(f"{file} already exists here")
        else:
            shutil.copyfile(os.path.join(package_dir, "templates", file), dest)
            typer.echo(f"Created {file} at {dest}")


def deploy_instructions():
    typer.echo("Printing deployment instructions")
    openai_api_key = os.environ.get("OPENAI_API_KEY", "your OpenAI API key")
    instructions = inspect.cleandoc(
        f"""
        # For fly.io:
        fly launch # answer no to Postgres, Redis and deploying now 
        fly secrets set OPENAI_API_KEY={openai_api_key} 
        fly deploy
        
        # For Google Cloud Run:
        gcloud run deploy --source . --set-env-vars="OPENAI_API_KEY={openai_api_key}"
        """
    )
    print(instructions)


def main(action: Optional[str] = typer.Argument(None)):
    if not action or action == "serve":
        preflight_checks()
        serve()
    elif action == "init":
        init()
    elif action == "deploy":
        preflight_checks()
        deploy_instructions()
    else:
        print(f"Unknown action: {action}")


# script entrypoint for flit
def cli_wrapper():
    typer.run(main)


if __name__ == "__main__":
    typer.run(main)
